from .introspection import JWTIntrospectionEndpoint as JWTIntrospectionEndpoint
from .revocation import JWTRevocationEndpoint as JWTRevocationEndpoint
from .token import JWTBearerTokenGenerator as JWTBearerTokenGenerator
from .token_validator import JWTBearerTokenValidator as JWTBearerTokenValidator

__all__ = ["JWTBearerTokenGenerator", "JWTBearerTokenValidator", "JWTIntrospectionEndpoint", "JWTRevocationEndpoint"]



from .models import AuthorizationServerMetadata as AuthorizationServerMetadata
from .well_known import get_well_known_url as get_well_known_url

__all__ = ["AuthorizationServerMetadata", "get_well_known_url"]



from __future__ import annotations

import asyncio
from typing import Awaitable, List, Tuple, Union
from typing_extensions import assert_type


async def coro1() -> int:
    return 42


async def coro2() -> str:
    return "spam"


async def test_gather(awaitable1: Awaitable[int], awaitable2: Awaitable[str]) -> None:
    a = await asyncio.gather(awaitable1)
    assert_type(a, Tuple[int])

    b = await asyncio.gather(awaitable1, awaitable2, return_exceptions=True)
    assert_type(b, Tuple[Union[int, BaseException], Union[str, BaseException]])

    c = await asyncio.gather(awaitable1, awaitable2, awaitable1, awaitable1, awaitable1, awaitable1)
    assert_type(c, Tuple[int, str, int, int, int, int])

    d = await asyncio.gather(awaitable1, awaitable1, awaitable1, awaitable1, awaitable1, awaitable1, awaitable1)
    assert_type(d, List[int])

    awaitables_list: list[Awaitable[int]] = [awaitable1]
    e = await asyncio.gather(*awaitables_list)
    assert_type(e, List[int])

    # this case isn't reliable between typecheckers, no one would ever call it with no args anyway
    # f = await asyncio.gather()
    # assert_type(f, list[Any])


asyncio.run(test_gather(coro1(), coro2()))



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import sys
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import tensorflow.core.protobuf.tpu.optimization_parameters_pb2

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class TPUEmbeddingConfiguration(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Mode:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ModeEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[TPUEmbeddingConfiguration._Mode.ValueType], builtins.type
    ):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNSPECIFIED: TPUEmbeddingConfiguration._Mode.ValueType  # 0
        INFERENCE: TPUEmbeddingConfiguration._Mode.ValueType  # 1
        TRAINING: TPUEmbeddingConfiguration._Mode.ValueType  # 2
        BACKWARD_PASS_ONLY: TPUEmbeddingConfiguration._Mode.ValueType  # 3

    class Mode(_Mode, metaclass=_ModeEnumTypeWrapper):
        """Mode. Should the embedding layer program be run for inference (just forward
        pass), training (both forward and backward pass) or just the backward_pass.
        """

    UNSPECIFIED: TPUEmbeddingConfiguration.Mode.ValueType  # 0
    INFERENCE: TPUEmbeddingConfiguration.Mode.ValueType  # 1
    TRAINING: TPUEmbeddingConfiguration.Mode.ValueType  # 2
    BACKWARD_PASS_ONLY: TPUEmbeddingConfiguration.Mode.ValueType  # 3

    class _ShardingStrategy:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ShardingStrategyEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[TPUEmbeddingConfiguration._ShardingStrategy.ValueType],
        builtins.type,
    ):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        DIV_DEFAULT: TPUEmbeddingConfiguration._ShardingStrategy.ValueType  # 0
        MOD: TPUEmbeddingConfiguration._ShardingStrategy.ValueType  # 1

    class ShardingStrategy(_ShardingStrategy, metaclass=_ShardingStrategyEnumTypeWrapper):
        """Sharding strategy of the embedding tables among the hosts.
        If the sharding_strategy is "mod", each id is assigned to host
        "id % num_hosts". For instance, 13 ids are split across 5 hosts as:
        [[0, 5, 10], [1, 6, 11], [2, 7, 12], [3, 8], [4, 9]].
        If the sharding_strategy is "div", ids are assigned to hosts in a
        contiguous manner. In this case, 13 ids are split across 5 hosts as:
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10], [11, 12]].
        In both the strategies, if the id space does not evenly divide the number
        of hosts, each of the first "table_descriptor.vocabulary_size % num_hosts"
        hosts will be assigned one more id.
        This partitioning strategy exactly follows that in the embedding_lookup
        TensorFlow function at tensorflow/python/ops/embedding_ops.py.
        """

    DIV_DEFAULT: TPUEmbeddingConfiguration.ShardingStrategy.ValueType  # 0
    MOD: TPUEmbeddingConfiguration.ShardingStrategy.ValueType  # 1

    @typing.final
    class TableDescriptor(google.protobuf.message.Message):
        """Description of the various embedding tables."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        VOCABULARY_SIZE_FIELD_NUMBER: builtins.int
        DIMENSION_FIELD_NUMBER: builtins.int
        NUM_FEATURES_FIELD_NUMBER: builtins.int
        OPTIMIZATION_PARAMETERS_FIELD_NUMBER: builtins.int
        name: builtins.str
        """Name of the table."""
        vocabulary_size: builtins.int
        """Size of the vocabulary (i.e., number of rows) in the table."""
        dimension: builtins.int
        """The embedding dimension (i.e., the width of the embedding table)."""
        num_features: builtins.int
        """Number of features mapped to this table."""
        @property
        def optimization_parameters(self) -> tensorflow.core.protobuf.tpu.optimization_parameters_pb2.OptimizationParameters:
            """Details of the learning algorithm used to update the embedding
            parameters.
            """

        def __init__(
            self,
            *,
            name: builtins.str | None = ...,
            vocabulary_size: builtins.int | None = ...,
            dimension: builtins.int | None = ...,
            num_features: builtins.int | None = ...,
            optimization_parameters: tensorflow.core.protobuf.tpu.optimization_parameters_pb2.OptimizationParameters | None = ...,
        ) -> None: ...
        def HasField(
            self, field_name: typing.Literal["optimization_parameters", b"optimization_parameters"]
        ) -> builtins.bool: ...
        def ClearField(
            self,
            field_name: typing.Literal[
                "dimension",
                b"dimension",
                "name",
                b"name",
                "num_features",
                b"num_features",
                "optimization_parameters",
                b"optimization_parameters",
                "vocabulary_size",
                b"vocabulary_size",
            ],
        ) -> None: ...

    @typing.final
    class FeatureDescriptor(google.protobuf.message.Message):
        """Description of different input features."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        TABLE_ID_FIELD_NUMBER: builtins.int
        INPUT_SHAPE_FIELD_NUMBER: builtins.int
        name: builtins.str
        """Name of the input feature."""
        table_id: builtins.int
        """Index of the corresponding table in the TableDescriptor list."""
        @property
        def input_shape(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
            """Static shape of the inputs (excluding the reduction axis). Note that
            the shape of the actual inputs provided using the infeed op must be
            strictly smaller than input_shape. The outputs received at the TensorCore
            will have rank = input_shape.size() + 1. The innermost axis corresponds
            to the embedding dimension. If the input has shape [m, n, k] (excluding
            the reduction axis) and the embedding dimension is d, the output received
            at the TensorCore will have shape [m, n, k, d].
            """

        def __init__(
            self,
            *,
            name: builtins.str | None = ...,
            table_id: builtins.int | None = ...,
            input_shape: collections.abc.Iterable[builtins.int] | None = ...,
        ) -> None: ...
        def ClearField(
            self, field_name: typing.Literal["input_shape", b"input_shape", "name", b"name", "table_id", b"table_id"]
        ) -> None: ...

    @typing.final
    class SpmdSharding(google.protobuf.message.Message):
        """SPMD (Single Program Multiple Data) sharding configuration for
        TPUEmbedding. When model parallelism is used on the TensorCore, the number
        of cores per replica must be passed to TPUEmbedding so that the right
        shapes can be computed in the TF/XLA bridge.
        """

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        ENABLED_FIELD_NUMBER: builtins.int
        NUM_CORES_PER_REPLICA_FIELD_NUMBER: builtins.int
        enabled: builtins.bool
        """Whether SPMD sharding is enabled."""
        num_cores_per_replica: builtins.int
        """Number of cores per replica."""
        def __init__(self, *, enabled: builtins.bool | None = ..., num_cores_per_replica: builtins.int | None = ...) -> None: ...
        def ClearField(
            self, field_name: typing.Literal["enabled", b"enabled", "num_cores_per_replica", b"num_cores_per_replica"]
        ) -> None: ...

    TABLE_DESCRIPTOR_FIELD_NUMBER: builtins.int
    MODE_FIELD_NUMBER: builtins.int
    BATCH_SIZE_PER_TENSOR_CORE_FIELD_NUMBER: builtins.int
    NUM_HOSTS_FIELD_NUMBER: builtins.int
    NUM_TENSOR_CORES_FIELD_NUMBER: builtins.int
    SHARDING_STRATEGY_FIELD_NUMBER: builtins.int
    PIPELINE_EXECUTION_WITH_TENSOR_CORE_FIELD_NUMBER: builtins.int
    PROFILE_DATA_DIRECTORY_FIELD_NUMBER: builtins.int
    FEATURE_DESCRIPTOR_FIELD_NUMBER: builtins.int
    SPMD_SHARDING_FIELD_NUMBER: builtins.int
    mode: global___TPUEmbeddingConfiguration.Mode.ValueType
    batch_size_per_tensor_core: builtins.int
    """Number of samples in each batch of embedding layer activations sent to
    the TensorCore.
    """
    num_hosts: builtins.int
    """Number of TPU hosts used for inference/training."""
    num_tensor_cores: builtins.int
    """Number of TensorCore used for inference/training."""
    sharding_strategy: global___TPUEmbeddingConfiguration.ShardingStrategy.ValueType
    pipeline_execution_with_tensor_core: builtins.bool
    """This parameter determines if the execution of the sparse core will be
    pipelined with that of the TensorCore. This parameter only affects results
    when mode=TRAINING. If mode=INFERENCE or BACKWARD_PASS_ONLY, this parameter
    does not affect execution and hence, is a don't care value.

    false: The execution of the sparse core is not pipelined with that of the
    TensorCore. The forward pass of every step on the sparse core is executed
    only after the backward pass of the previous step is complete. And the
    backward pass on the sparse core is executed only after the embedding
    gradients have been computed on the TensorCore on every step. This ensures
    that the activations on every step observe the gradient updates from the
    previous step on both the sparse core and the TensorCore.

    true: The execution of the sparse core is pipelined with that of the
    TensorCore. The forward pass of every step on the sparse core can be
    executed after the forward pass of the previous step is complete without
    waiting for the backward pass. This improves the utilization of the sparse
    core allowing it to process step N+1 while the embedding gradients for step
    N are computed on the TensorCore. The backward pass of every step on the
    sparse core is executed directly after the forward pass for the next step
    is complete. The drawback is that embedding activations for step N+1 do not
    observe the embedding gradient updates from step N. This could affect model
    quality if step N and N+1 involve the same set of embedding IDs. However,
    since the embedding updates are sparse, this is generally not considered a
    problem.
    """
    profile_data_directory: builtins.str
    """Directory where embedding lookup statistics are stored. These statistics
    summarize information about the inputs to the embedding lookup
    operation, in particular, the average number of embedding IDs per example
    and how well the embedding IDs are load balanced across the system. The
    lookup statistics are used during TPU initialization for embedding table
    partitioning. Collection of lookup statistics is done at runtime by
    profiling the embedding inputs: only 3% of input samples are profiled to
    minimize host CPU overhead. Once a suitable number of samples are
    profiled, the lookup statistics are saved to table-specific files in the
    profile data directory generally at the end of a TPU training loop. The
    filename corresponding to each table is obtained by hashing table specific
    parameters (e.g., table name and number of features) and global
    configuration parameters (e.g., sharding strategy and TPU worker task
    count). The same profile data directory can be shared amongst several
    models to reuse embedding lookup statistics.
    """
    @property
    def table_descriptor(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___TPUEmbeddingConfiguration.TableDescriptor
    ]: ...
    @property
    def feature_descriptor(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___TPUEmbeddingConfiguration.FeatureDescriptor
    ]:
        """If the feature_descriptor field is populated, the model should NOT populate
        TableDescriptor.num_features and batch_size_per_tensor_core. These two
        fields will be auto-populated by the TPUEmbedding rewrite passes.
        """

    @property
    def spmd_sharding(self) -> global___TPUEmbeddingConfiguration.SpmdSharding: ...
    def __init__(
        self,
        *,
        table_descriptor: collections.abc.Iterable[global___TPUEmbeddingConfiguration.TableDescriptor] | None = ...,
        mode: global___TPUEmbeddingConfiguration.Mode.ValueType | None = ...,
        batch_size_per_tensor_core: builtins.int | None = ...,
        num_hosts: builtins.int | None = ...,
        num_tensor_cores: builtins.int | None = ...,
        sharding_strategy: global___TPUEmbeddingConfiguration.ShardingStrategy.ValueType | None = ...,
        pipeline_execution_with_tensor_core: builtins.bool | None = ...,
        profile_data_directory: builtins.str | None = ...,
        feature_descriptor: collections.abc.Iterable[global___TPUEmbeddingConfiguration.FeatureDescriptor] | None = ...,
        spmd_sharding: global___TPUEmbeddingConfiguration.SpmdSharding | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["spmd_sharding", b"spmd_sharding"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "batch_size_per_tensor_core",
            b"batch_size_per_tensor_core",
            "feature_descriptor",
            b"feature_descriptor",
            "mode",
            b"mode",
            "num_hosts",
            b"num_hosts",
            "num_tensor_cores",
            b"num_tensor_cores",
            "pipeline_execution_with_tensor_core",
            b"pipeline_execution_with_tensor_core",
            "profile_data_directory",
            b"profile_data_directory",
            "sharding_strategy",
            b"sharding_strategy",
            "spmd_sharding",
            b"spmd_sharding",
            "table_descriptor",
            b"table_descriptor",
        ],
    ) -> None: ...

global___TPUEmbeddingConfiguration = TPUEmbeddingConfiguration

@typing.final
class TPUEmbeddingError(google.protobuf.message.Message):
    """A placeholder message that is used to define a unique Status payload
    URL for TPU embedding errors.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___TPUEmbeddingError = TPUEmbeddingError



from _typeshed import Unused
from typing import Final, Literal

from Xlib.display import Display
from Xlib.protocol import rq
from Xlib.xobject import resource

extname: Final = "DPMS"
DPMSModeOn: Final = 0
DPMSModeStandby: Final = 1
DPMSModeSuspend: Final = 2
DPMSModeOff: Final = 3
DPMSPowerLevel: Final[tuple[Literal[0], Literal[1], Literal[2], Literal[3]]]

class DPMSGetVersion(rq.ReplyRequest): ...

def get_version(self: Display | resource.Resource) -> DPMSGetVersion: ...

class DPMSCapable(rq.ReplyRequest): ...

def capable(self: Display | resource.Resource) -> DPMSCapable: ...

class DPMSGetTimeouts(rq.ReplyRequest): ...

def get_timeouts(self: Display | resource.Resource) -> DPMSGetTimeouts: ...

class DPMSSetTimeouts(rq.Request): ...

def set_timeouts(
    self: Display | resource.Resource, standby_timeout: int, suspend_timeout: int, off_timeout: int
) -> DPMSSetTimeouts: ...

class DPMSEnable(rq.Request): ...

def enable(self: Display | resource.Resource) -> DPMSEnable: ...

class DPMSDisable(rq.Request): ...

def disable(self: Display | resource.Resource) -> DPMSDisable: ...

class DPMSForceLevel(rq.Request): ...

def force_level(self: Display | resource.Resource, power_level: int) -> DPMSForceLevel: ...

class DPMSInfo(rq.ReplyRequest): ...

def info(self: Display | resource.Resource) -> DPMSInfo: ...
def init(disp: Display, _info: Unused) -> None: ...



from _typeshed import Incomplete, Unused
from collections.abc import Callable
from typing import Final, Literal
from typing_extensions import TypeAlias

from reportlab.lib.colors import Color
from reportlab.pdfbase.ttfonts import ShapedStr
from reportlab.pdfgen.canvas import Canvas

# NOTE: This is slightly different from what toColor accepts and interprets
_Color: TypeAlias = Color | tuple[float, float, float, float] | tuple[float, float, float] | list[float] | str

__version__: Final[str]
log2vis: Callable[..., str | None]
BidiStr: type[str]
BidiList: type[list[Incomplete]]
BidiIndex: Incomplete

def bidiText(text: str, direction: str | None) -> str: ...
def bidiShapedText(
    text: str, direction: str = "RTL", clean: bool = True, fontName: str = "Helvetica", fontSize: int = 10, shaping: bool = False
) -> tuple[ShapedStr | str, float]: ...
def isBidiStr(_: Unused) -> bool: ...
def isBidiList(_: Unused) -> bool: ...
def innerBidiStrWrap(s: str, bidiV: int = -1, bidiL: int = -1) -> str: ...
def bidiStrWrap(s: str, orig: str) -> str: ...
def bidiListWrap(L, orig) -> list[Incomplete]: ...
def bidiFragWord(w: str, direction: str | None = None, bidiV: int = -1, bidiL: int = -1, clean: bool = True): ...
def bidiWordList(
    words: list[str] | tuple[str], direction: str = "RTL", clean: bool = True, wx: bool = False
) -> list[Incomplete]: ...

rtlSupport: bool

class _PDFColorSetter:
    def setFillColorCMYK(self, c: float, m: float, y: float, k: float, alpha: float | None = None) -> None: ...
    def setStrokeColorCMYK(self, c: float, m: float, y: float, k: float, alpha: float | None = None) -> None: ...
    def setFillColorRGB(self, r: float, g: float, b: float, alpha: float | None = None) -> None: ...
    def setStrokeColorRGB(self, r: float, g: float, b: float, alpha: float | None = None) -> None: ...
    def setFillColor(self, aColor: _Color, alpha: float | None = None) -> None: ...
    def setStrokeColor(self, aColor: _Color, alpha: float | None = None) -> None: ...
    def setFillGray(self, gray: float, alpha: float | None = None) -> None: ...
    def setStrokeGray(self, gray: float, alpha: float | None = None) -> None: ...
    def setStrokeAlpha(self, a: float) -> None: ...
    def setFillAlpha(self, a: float) -> None: ...
    def setStrokeOverprint(self, a) -> None: ...
    def setFillOverprint(self, a) -> None: ...
    def setOverprintMask(self, a) -> None: ...

class PDFTextObject(_PDFColorSetter):
    direction: Literal["LTR", "RTL"]
    def __init__(self, canvas: Canvas, x: float = 0, y: float = 0, direction: Literal["LTR", "RTL"] | None = None) -> None: ...
    def getCode(self) -> str: ...
    def setTextOrigin(self, x: float, y: float) -> None: ...
    def setTextTransform(self, a: float, b: float, c: float, d: float, e: float, f: float) -> None: ...
    def moveCursor(self, dx: float, dy: float) -> None: ...
    def setXPos(self, dx: float) -> None: ...
    def getCursor(self) -> tuple[float, float]: ...
    def getStartOfLine(self) -> tuple[float, float]: ...
    def getX(self) -> float: ...
    def getY(self) -> float: ...
    def setFont(self, psfontname: str, size: float, leading: float | None = None) -> None: ...
    def setCharSpace(self, charSpace: float) -> None: ...
    def setWordSpace(self, wordSpace: float) -> None: ...
    def setHorizScale(self, horizScale: float) -> None: ...
    def setLeading(self, leading: float) -> None: ...
    def setTextRenderMode(self, mode: Literal[0, 1, 2, 3, 4, 5, 6, 7]) -> None: ...
    def setRise(self, rise: float) -> None: ...
    def textOut(self, text: str) -> None: ...
    def textLine(self, text: str = "") -> None: ...
    def textLines(self, stuff: list[str] | tuple[str, ...] | str, trim: int = 1) -> None: ...
    def __nonzero__(self) -> bool: ...



from _typeshed import Incomplete
from typing import Final

from reportlab.platypus import Flowable

__version__: Final[str]
captionStyle: Incomplete

class Figure(Flowable):
    width: Incomplete
    figureHeight: Incomplete
    caption: Incomplete
    captionFont: Incomplete
    captionSize: Incomplete
    captionTextColor: Incomplete
    captionBackColor: Incomplete
    captionGap: Incomplete
    captionAlign: Incomplete
    captionPosition: Incomplete
    captionHeight: int
    background: Incomplete
    border: Incomplete
    spaceBefore: Incomplete
    spaceAfter: Incomplete
    hAlign: Incomplete
    def __init__(
        self,
        width,
        height,
        caption: str = "",
        captionFont="Helvetica-Oblique",
        captionSize: int = 12,
        background=None,
        captionTextColor=...,
        captionBackColor=None,
        border=None,
        spaceBefore: int = 12,
        spaceAfter: int = 12,
        captionGap=None,
        captionAlign: str = "centre",
        captionPosition: str = "bottom",
        hAlign: str = "CENTER",
    ) -> None: ...
    height: Incomplete
    dx: Incomplete
    def wrap(self, availWidth, availHeight): ...
    def draw(self) -> None: ...
    def drawBorder(self) -> None: ...
    def drawBackground(self) -> None: ...
    def drawCaption(self) -> None: ...
    def drawFigure(self) -> None: ...

def drawPage(canvas, x, y, width, height) -> None: ...

class PageFigure(Figure):
    caption: str
    captionStyle: Incomplete
    background: Incomplete
    def __init__(self, background=None) -> None: ...
    def drawVirtualPage(self) -> None: ...
    def drawFigure(self) -> None: ...

class PlatPropFigure1(PageFigure):
    caption: str
    def __init__(self) -> None: ...
    def drawVirtualPage(self) -> None: ...

class FlexFigure(Figure):
    shrinkToFit: Incomplete
    growToFit: Incomplete
    scaleFactor: Incomplete
    background: Incomplete
    def __init__(
        self,
        width,
        height,
        caption,
        background=None,
        captionFont: str = "Helvetica-Oblique",
        captionSize: int = 8,
        captionTextColor=...,
        shrinkToFit: int = 1,
        growToFit: int = 1,
        spaceBefore: int = 12,
        spaceAfter: int = 12,
        captionGap: int = 9,
        captionAlign: str = "centre",
        captionPosition: str = "top",
        scaleFactor=None,
        hAlign: str = "CENTER",
        border: int = 1,
    ) -> None: ...
    def wrap(self, availWidth, availHeight): ...
    def split(self, availWidth, availHeight): ...

class ImageFigure(FlexFigure):
    filename: Incomplete
    def __init__(self, filename, caption, background=None, scaleFactor=None, hAlign: str = "CENTER", border=None) -> None: ...
    def drawFigure(self) -> None: ...

class DrawingFigure(FlexFigure):
    drawing: Incomplete
    growToFit: int
    def __init__(self, modulename, classname, caption, baseDir=None, background=None) -> None: ...
    def drawFigure(self) -> None: ...

def demo1(canvas) -> None: ...
def test1() -> None: ...



class ISAPIError(Exception):
    errno: int
    strerror: str | None
    funcname: str | None
    def __init__(self, errno: int, strerror: str | None = None, funcname: str | None = None) -> None: ...

class FilterError(ISAPIError): ...
class ExtensionError(ISAPIError): ...
class InternalReloadException(Exception): ...



from logging import Logger

from .errors import (
    AuthorizationPendingError as AuthorizationPendingError,
    ExpiredTokenError as ExpiredTokenError,
    SlowDownError as SlowDownError,
)

log: Logger



from hvac.api.auth_methods.jwt import JWT

class OIDC(JWT):
    DEFAULT_PATH: str
    def create_role(
        self,
        name,
        user_claim,
        allowed_redirect_uris,
        role_type: str = "oidc",
        bound_audiences=None,
        clock_skew_leeway=None,
        expiration_leeway=None,
        not_before_leeway=None,
        bound_subject=None,
        bound_claims=None,
        groups_claim=None,
        claim_mappings=None,
        oidc_scopes=None,
        bound_claims_type: str = "string",
        verbose_oidc_logging: bool = False,
        token_ttl=None,
        token_max_ttl=None,
        token_policies=None,
        token_bound_cidrs=None,
        token_explicit_max_ttl=None,
        token_no_default_policy=None,
        token_num_uses=None,
        token_period=None,
        token_type=None,
        path=None,
        user_claim_json_pointer=None,
    ) -> None: ...



from collections.abc import Callable, Iterable
from typing import Any
from typing_extensions import TypeAlias

import tensorflow as tf
from tensorflow._aliases import Gradients
from tensorflow.keras.optimizers import schedules as schedules
from tensorflow.python.trackable.base import Trackable

_LearningRate: TypeAlias = float | tf.Tensor | schedules.LearningRateSchedule | Callable[[], float | tf.Tensor]
_GradientAggregator: TypeAlias = Callable[[list[tuple[Gradients, tf.Variable]]], list[tuple[Gradients, tf.Variable]]] | None
_GradientTransformer: TypeAlias = (
    Iterable[Callable[[list[tuple[Gradients, tf.Variable]]], list[tuple[Gradients, tf.Variable]]]] | None
)

# kwargs here and in other optimizers can be given better type after Unpack[TypedDict], PEP 692, is supported.
class Optimizer(Trackable):
    _name: str
    _iterations: tf.Variable | None
    _weights: list[tf.Variable]
    gradient_aggregator: _GradientAggregator
    gradient_transformers: _GradientTransformer
    learning_rate: _LearningRate
    def __init__(
        self,
        name: str,
        gradient_aggregator: _GradientAggregator = None,
        gradient_transformers: _GradientTransformer = None,
        **kwargs: Any,
    ) -> None: ...

class Adam(Optimizer):
    def __init__(
        self,
        learning_rate: _LearningRate = 0.001,
        beta_1: float = 0.9,
        beta_2: float = 0.999,
        epsilon: float = 1e-07,
        amsgrad: bool = False,
        name: str = "Adam",
        **kwargs: Any,
    ) -> None: ...

class Adagrad(Optimizer):
    _initial_accumulator_value: float

    def __init__(
        self,
        learning_rate: _LearningRate = 0.001,
        initial_accumulator_value: float = 0.1,
        epsilon: float = 1e-7,
        name: str = "Adagrad",
        **kwargs: Any,
    ) -> None: ...

class SGD(Optimizer):
    def __init__(
        self, learning_rate: _LearningRate = 0.01, momentum: float = 0.0, nesterov: bool = False, name: str = "SGD", **kwargs: Any
    ) -> None: ...

def __getattr__(name: str): ...  # incomplete module



from _typeshed import SupportsRead
from typing import Any, Final, Literal, NoReturn
from xml.dom.minidom import Document, Node, _DOMErrorHandler

__all__ = ["DOMBuilder", "DOMEntityResolver", "DOMInputSource"]

class Options:
    namespaces: int
    namespace_declarations: bool
    validation: bool
    external_parameter_entities: bool
    external_general_entities: bool
    external_dtd_subset: bool
    validate_if_schema: bool
    validate: bool
    datatype_normalization: bool
    create_entity_ref_nodes: bool
    entities: bool
    whitespace_in_element_content: bool
    cdata_sections: bool
    comments: bool
    charset_overrides_xml_encoding: bool
    infoset: bool
    supported_mediatypes_only: bool
    errorHandler: _DOMErrorHandler | None
    filter: DOMBuilderFilter | None

class DOMBuilder:
    entityResolver: DOMEntityResolver | None
    errorHandler: _DOMErrorHandler | None
    filter: DOMBuilderFilter | None
    ACTION_REPLACE: Final = 1
    ACTION_APPEND_AS_CHILDREN: Final = 2
    ACTION_INSERT_AFTER: Final = 3
    ACTION_INSERT_BEFORE: Final = 4
    def __init__(self) -> None: ...
    def setFeature(self, name: str, state: int) -> None: ...
    def supportsFeature(self, name: str) -> bool: ...
    def canSetFeature(self, name: str, state: Literal[1, 0]) -> bool: ...
    # getFeature could return any attribute from an instance of `Options`
    def getFeature(self, name: str) -> Any: ...
    def parseURI(self, uri: str) -> Document: ...
    def parse(self, input: DOMInputSource) -> Document: ...
    def parseWithContext(self, input: DOMInputSource, cnode: Node, action: Literal[1, 2, 3, 4]) -> NoReturn: ...

class DOMEntityResolver:
    __slots__ = ("_opener",)
    def resolveEntity(self, publicId: str | None, systemId: str) -> DOMInputSource: ...

class DOMInputSource:
    __slots__ = ("byteStream", "characterStream", "stringData", "encoding", "publicId", "systemId", "baseURI")
    byteStream: SupportsRead[bytes] | None
    characterStream: SupportsRead[str] | None
    stringData: str | None
    encoding: str | None
    publicId: str | None
    systemId: str | None
    baseURI: str | None

class DOMBuilderFilter:
    FILTER_ACCEPT: Final = 1
    FILTER_REJECT: Final = 2
    FILTER_SKIP: Final = 3
    FILTER_INTERRUPT: Final = 4
    whatToShow: int
    def acceptNode(self, element: Node) -> Literal[1, 2, 3, 4]: ...
    def startContainer(self, element: Node) -> Literal[1, 2, 3, 4]: ...

class DocumentLS:
    async_: bool
    def abort(self) -> NoReturn: ...
    def load(self, uri: str) -> NoReturn: ...
    def loadXML(self, source: str) -> NoReturn: ...
    def saveXML(self, snode: Node | None) -> str: ...

class DOMImplementationLS:
    MODE_SYNCHRONOUS: Final = 1
    MODE_ASYNCHRONOUS: Final = 2
    def createDOMBuilder(self, mode: Literal[1], schemaType: None) -> DOMBuilder: ...
    def createDOMWriter(self) -> NoReturn: ...
    def createDOMInputSource(self) -> DOMInputSource: ...



from _typeshed import SupportsRead
from typing import Any
from typing_extensions import TypeAlias

from yaml.error import YAMLError

_ReadStream: TypeAlias = str | bytes | SupportsRead[str] | SupportsRead[bytes]

class ReaderError(YAMLError):
    name: Any
    character: Any
    position: Any
    encoding: Any
    reason: Any
    def __init__(self, name, position, character, encoding, reason) -> None: ...

class Reader:
    name: Any
    stream: SupportsRead[str] | SupportsRead[bytes] | None
    stream_pointer: Any
    eof: Any
    buffer: Any
    pointer: Any
    raw_buffer: Any
    raw_decode: Any
    encoding: Any
    index: Any
    line: Any
    column: Any
    def __init__(self, stream: _ReadStream) -> None: ...
    def peek(self, index=0): ...
    def prefix(self, length=1): ...
    def forward(self, length=1): ...
    def get_mark(self): ...
    def determine_encoding(self): ...
    NON_PRINTABLE: Any
    def check_printable(self, data): ...
    def update(self, length): ...
    def update_raw(self, size=4096): ...

__all__ = ["Reader", "ReaderError"]



from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["clique_removal", "max_clique", "large_clique_size", "maximum_independent_set"]

@_dispatchable
def maximum_independent_set(G: Graph[_Node]) -> set[_Node]: ...
@_dispatchable
def max_clique(G: Graph[_Node]) -> set[_Node]: ...
@_dispatchable
def clique_removal(G: Graph[_Node]) -> tuple[set[_Node], list[set[_Node]]]: ...
@_dispatchable
def large_clique_size(G: Graph[_Node]) -> int: ...



from collections.abc import Iterable, Sequence
from typing_extensions import TypeAlias

from Xlib.protocol import rq

# Aliases used in other modules
_RGB3IntIterable: TypeAlias = Iterable[int]  # noqa: Y047
_Rectangle4IntSequence: TypeAlias = Sequence[int]  # noqa: Y047
_Segment4IntSequence: TypeAlias = Sequence[int]  # noqa: Y047
_Arc6IntSequence: TypeAlias = Sequence[int]  # noqa: Y047

# TODO: Complete all classes using WindowValues and GCValues
# Currently *object is used to represent the ValueList instead of the possible attribute types
def WindowValues(arg: str) -> rq.ValueList: ...
def GCValues(arg: str) -> rq.ValueList: ...

TimeCoord: rq.Struct
Host: rq.Struct
CharInfo: rq.Struct
FontProp: rq.Struct
ColorItem: rq.Struct
RGB: rq.Struct
Point: rq.Struct
Segment: rq.Struct
Rectangle: rq.Struct
Arc: rq.Struct



import abc
import enum
import threading
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from concurrent import futures
from types import ModuleType, TracebackType
from typing import Any, Generic, NoReturn, Protocol, TypeVar, type_check_only
from typing_extensions import Self, TypeAlias

from . import aio as aio

__version__: str

_T = TypeVar("_T")

# XXX: Early attempts to tame this used literals for all the keys (gRPC is
# a bit segfaulty and doesn't adequately validate the option keys), but that
# didn't quite work out. Maybe it's something we can come back to
_OptionKeyValue: TypeAlias = tuple[str, Any]
_Options: TypeAlias = Sequence[_OptionKeyValue]

class Compression(enum.IntEnum):
    NoCompression = 0
    Deflate = 1
    Gzip = 2

@enum.unique
class LocalConnectionType(enum.Enum):
    UDS = 0
    LOCAL_TCP = 1

# XXX: not documented, needs more investigation.
# Some evidence:
# - https://github.com/grpc/grpc/blob/0e1984effd7e977ef18f1ad7fde7d10a2a153e1d/src/python/grpcio_tests/tests/unit/_metadata_test.py#L71
# - https://github.com/grpc/grpc/blob/0e1984effd7e977ef18f1ad7fde7d10a2a153e1d/src/python/grpcio_tests/tests/unit/_metadata_test.py#L58
# - https://github.com/grpc/grpc/blob/0e1984effd7e977ef18f1ad7fde7d10a2a153e1d/src/python/grpcio_tests/tests/unit/_invocation_defects_test.py#L66
_Metadata: TypeAlias = tuple[tuple[str, str | bytes], ...]

_TRequest = TypeVar("_TRequest")
_TResponse = TypeVar("_TResponse")
_Serializer: TypeAlias = Callable[[_T], bytes]
_Deserializer: TypeAlias = Callable[[bytes], _T]

# Future Interfaces:

class FutureTimeoutError(Exception): ...
class FutureCancelledError(Exception): ...

_TFutureValue = TypeVar("_TFutureValue")

class Future(abc.ABC, Generic[_TFutureValue]):
    @abc.abstractmethod
    def add_done_callback(self, fn: Callable[[Future[_TFutureValue]], None]) -> None: ...
    @abc.abstractmethod
    def cancel(self) -> bool: ...
    @abc.abstractmethod
    def cancelled(self) -> bool: ...
    @abc.abstractmethod
    def done(self) -> bool: ...
    @abc.abstractmethod
    def exception(self, timeout: float | None = None) -> Exception | None: ...
    @abc.abstractmethod
    def result(self, timeout: float | None = None) -> _TFutureValue: ...
    @abc.abstractmethod
    def running(self) -> bool: ...

    # FIXME: unsure of the exact return type here. Is it a traceback.StackSummary?
    @abc.abstractmethod
    def traceback(self, timeout: float | None = None): ...

# Create Client:

def insecure_channel(target: str, options: _Options | None = None, compression: Compression | None = None) -> Channel: ...
def secure_channel(
    target: str, credentials: ChannelCredentials, options: _Options | None = None, compression: Compression | None = None
) -> Channel: ...

_Interceptor: TypeAlias = (
    UnaryUnaryClientInterceptor | UnaryStreamClientInterceptor | StreamUnaryClientInterceptor | StreamStreamClientInterceptor
)

def intercept_channel(channel: Channel, *interceptors: _Interceptor) -> Channel: ...

# Create Client Credentials:

def ssl_channel_credentials(
    root_certificates: bytes | None = None, private_key: bytes | None = None, certificate_chain: bytes | None = None
) -> ChannelCredentials: ...
def local_channel_credentials(local_connect_type: LocalConnectionType = ...) -> ChannelCredentials: ...
def metadata_call_credentials(metadata_plugin: AuthMetadataPlugin, name: str | None = None) -> CallCredentials: ...
def access_token_call_credentials(access_token: str) -> CallCredentials: ...
def alts_channel_credentials(service_accounts: Sequence[str] | None = None) -> ChannelCredentials: ...
def compute_engine_channel_credentials(call_credentials: CallCredentials) -> ChannelCredentials: ...
def xds_channel_credentials(fallback_credentials: ChannelCredentials | None = None) -> ChannelCredentials: ...

# GRPC docs say there should be at least two:
def composite_call_credentials(creds1: CallCredentials, creds2: CallCredentials, *rest: CallCredentials) -> CallCredentials: ...

# Compose a ChannelCredentials and one or more CallCredentials objects.
def composite_channel_credentials(
    channel_credentials: ChannelCredentials, call_credentials: CallCredentials, *rest: CallCredentials
) -> ChannelCredentials: ...

# Create Server:

def server(
    thread_pool: futures.ThreadPoolExecutor,
    handlers: list[GenericRpcHandler] | None = None,
    interceptors: list[ServerInterceptor] | None = None,
    options: _Options | None = None,
    maximum_concurrent_rpcs: int | None = None,
    compression: Compression | None = None,
    xds: bool = False,
) -> Server: ...

# Create Server Credentials:

_CertificateChainPair: TypeAlias = tuple[bytes, bytes]

def ssl_server_credentials(
    private_key_certificate_chain_pairs: list[_CertificateChainPair],
    root_certificates: bytes | None = None,
    require_client_auth: bool = False,
) -> ServerCredentials: ...
def local_server_credentials(local_connect_type: LocalConnectionType = ...) -> ServerCredentials: ...
def ssl_server_certificate_configuration(
    private_key_certificate_chain_pairs: list[_CertificateChainPair], root_certificates: bytes | None = None
) -> ServerCertificateConfiguration: ...
def dynamic_ssl_server_credentials(
    initial_certificate_configuration: ServerCertificateConfiguration,
    certificate_configuration_fetcher: Callable[[], ServerCertificateConfiguration],
    require_client_authentication: bool = False,
) -> ServerCredentials: ...
def alts_server_credentials() -> ServerCredentials: ...
def insecure_server_credentials() -> ServerCredentials: ...
def xds_server_credentials(fallback_credentials: ServerCredentials) -> ServerCredentials: ...

# RPC Method Handlers:

# XXX: This is probably what appears in the add_FooServicer_to_server function
# in the _pb2_grpc files that get generated, which points to the FooServicer
# handler functions that get generated, which look like this:
#
#    def FloobDoob(self, request, context):
#       return response
#
@type_check_only
class _Behaviour(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

def unary_unary_rpc_method_handler(
    behavior: _Behaviour,
    request_deserializer: _Deserializer[_TRequest] | None = None,
    response_serializer: _Serializer[_TResponse] | None = None,
) -> RpcMethodHandler[_TRequest, _TResponse]: ...
def unary_stream_rpc_method_handler(
    behavior: _Behaviour,
    request_deserializer: _Deserializer[_TRequest] | None = None,
    response_serializer: _Serializer[_TResponse] | None = None,
) -> RpcMethodHandler[_TRequest, _TResponse]: ...
def stream_unary_rpc_method_handler(
    behavior: _Behaviour,
    request_deserializer: _Deserializer[_TRequest] | None = None,
    response_serializer: _Serializer[_TResponse] | None = None,
) -> RpcMethodHandler[_TRequest, _TResponse]: ...
def stream_stream_rpc_method_handler(
    behavior: _Behaviour,
    request_deserializer: _Deserializer[_TRequest] | None = None,
    response_serializer: _Serializer[_TResponse] | None = None,
) -> RpcMethodHandler[_TRequest, _TResponse]: ...
def method_handlers_generic_handler(
    service: str, method_handlers: dict[str, RpcMethodHandler[Any, Any]]
) -> GenericRpcHandler: ...

# Channel Ready Future:

def channel_ready_future(channel: Channel) -> Future[None]: ...

# Channel Connectivity:

class ChannelConnectivity(enum.Enum):
    IDLE = (0, "idle")
    CONNECTING = (1, "connecting")
    READY = (2, "ready")
    TRANSIENT_FAILURE = (3, "transient failure")
    SHUTDOWN = (4, "shutdown")

# gRPC Status Code:

class Status(abc.ABC):
    code: StatusCode

    # XXX: misnamed property, does not align with status.proto, where it is called 'message':
    details: str

    trailing_metadata: _Metadata

# https://grpc.github.io/grpc/core/md_doc_statuscodes.html
class StatusCode(enum.Enum):
    OK = (0, "ok")
    CANCELLED = (1, "cancelled")
    UNKNOWN = (2, "unknown")
    INVALID_ARGUMENT = (3, "invalid argument")
    DEADLINE_EXCEEDED = (4, "deadline exceeded")
    NOT_FOUND = (5, "not found")
    ALREADY_EXISTS = (6, "already exists")
    PERMISSION_DENIED = (7, "permission denied")
    RESOURCE_EXHAUSTED = (8, "resource exhausted")
    FAILED_PRECONDITION = (9, "failed precondition")
    ABORTED = (10, "aborted")
    OUT_OF_RANGE = (11, "out of range")
    UNIMPLEMENTED = (12, "unimplemented")
    INTERNAL = (13, "internal")
    UNAVAILABLE = (14, "unavailable")
    DATA_LOSS = (15, "data loss")
    UNAUTHENTICATED = (16, "unauthenticated")

# Channel Object:

class Channel(abc.ABC):
    @abc.abstractmethod
    def close(self) -> None: ...
    @abc.abstractmethod
    def stream_stream(
        self,
        method: str,
        request_serializer: _Serializer[_TRequest] | None = None,
        response_deserializer: _Deserializer[_TResponse] | None = None,
    ) -> StreamStreamMultiCallable[_TRequest, _TResponse]: ...
    @abc.abstractmethod
    def stream_unary(
        self,
        method: str,
        request_serializer: _Serializer[_TRequest] | None = None,
        response_deserializer: _Deserializer[_TResponse] | None = None,
    ) -> StreamUnaryMultiCallable[_TRequest, _TResponse]: ...
    @abc.abstractmethod
    def subscribe(self, callback: Callable[[ChannelConnectivity], None], try_to_connect: bool = False) -> None: ...
    @abc.abstractmethod
    def unary_stream(
        self,
        method: str,
        request_serializer: _Serializer[_TRequest] | None = None,
        response_deserializer: _Deserializer[_TResponse] | None = None,
    ) -> UnaryStreamMultiCallable[_TRequest, _TResponse]: ...
    @abc.abstractmethod
    def unary_unary(
        self,
        method: str,
        request_serializer: _Serializer[_TRequest] | None = None,
        response_deserializer: _Deserializer[_TResponse] | None = None,
    ) -> UnaryUnaryMultiCallable[_TRequest, _TResponse]: ...
    @abc.abstractmethod
    def unsubscribe(self, callback: Callable[[ChannelConnectivity], None]) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None
    ) -> bool | None: ...

# Server Object:

class Server(abc.ABC):
    @abc.abstractmethod
    def add_generic_rpc_handlers(self, generic_rpc_handlers: Iterable[GenericRpcHandler]) -> None: ...

    # Returns an integer port on which server will accept RPC requests.
    @abc.abstractmethod
    def add_insecure_port(self, address: str) -> int: ...

    # Returns an integer port on which server will accept RPC requests.
    @abc.abstractmethod
    def add_secure_port(self, address: str, server_credentials: ServerCredentials) -> int: ...
    @abc.abstractmethod
    def start(self) -> None: ...

    # Grace period is in seconds.
    @abc.abstractmethod
    def stop(self, grace: float | None) -> threading.Event: ...

    # Block current thread until the server stops. Returns a bool
    # indicates if the operation times out. Timeout is in seconds.
    def wait_for_termination(self, timeout: float | None = None) -> bool: ...

# Authentication & Authorization Objects:

# This class has no supported interface
class ChannelCredentials:
    def __init__(self, credentials) -> None: ...

# This class has no supported interface
class CallCredentials:
    def __init__(self, credentials) -> None: ...

class AuthMetadataContext(abc.ABC):
    service_url: str
    method_name: str

class AuthMetadataPluginCallback(abc.ABC):
    def __call__(self, metadata: _Metadata, error: Exception | None) -> None: ...

class AuthMetadataPlugin(abc.ABC):
    def __call__(self, context: AuthMetadataContext, callback: AuthMetadataPluginCallback) -> None: ...

# This class has no supported interface
class ServerCredentials:
    def __init__(self, credentials) -> None: ...

# This class has no supported interface
class ServerCertificateConfiguration:
    def __init__(self, certificate_configuration) -> None: ...

# gRPC Exceptions:

@type_check_only
class _Metadatum:
    key: str
    value: bytes

# FIXME: There is scant documentation about what is actually available in this type.
# The properties here are the properties observed in the wild, and may be inaccurate.
# A better source to confirm their presence needs to be found at some point.
class RpcError(Exception):
    def code(self) -> StatusCode: ...

    # misnamed property, does not align with status.proto, where it is called 'message':
    def details(self) -> str | None: ...

    # XXX: This has a slightly different return type to all the other metadata:
    def trailing_metadata(self) -> tuple[_Metadatum, ...]: ...

# Shared Context:

class RpcContext(abc.ABC):
    @abc.abstractmethod
    def add_callback(self, callback: Callable[[], None]) -> bool: ...
    @abc.abstractmethod
    def cancel(self) -> bool: ...
    @abc.abstractmethod
    def is_active(self) -> bool: ...
    @abc.abstractmethod
    def time_remaining(self) -> float: ...

# Client-Side Context:

class Call(RpcContext, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def code(self) -> StatusCode: ...

    # misnamed property, does not align with status.proto, where it is called 'message':
    @abc.abstractmethod
    def details(self) -> str: ...
    @abc.abstractmethod
    def initial_metadata(self) -> _Metadata: ...
    @abc.abstractmethod
    def trailing_metadata(self) -> _Metadata: ...

# Client-Side Interceptor:

class ClientCallDetails(abc.ABC):
    method: str
    timeout: float | None
    metadata: _Metadata | None
    credentials: CallCredentials | None

    # "This is an EXPERIMENTAL argument. An optional flag t enable wait for ready mechanism."
    wait_for_ready: bool | None

    compression: Compression | None

# An object that is both a Call for the RPC and a Future. In the event of
# RPC completion, the return Call-Future's result value will be the
# response message of the RPC. Should the event terminate with non-OK
# status, the returned Call-Future's exception value will be an RpcError.
#
@type_check_only
class _CallFuture(Call, Future[_TResponse], metaclass=abc.ABCMeta): ...

class UnaryUnaryClientInterceptor(abc.ABC):
    # This method (not the class) is generic over _TRequest and _TResponse
    # and the types must satisfy the no-op implementation of
    # `return continuation(client_call_details, request)`.
    @abc.abstractmethod
    def intercept_unary_unary(
        self,
        continuation: Callable[[ClientCallDetails, _TRequest], _CallFuture[_TResponse]],
        client_call_details: ClientCallDetails,
        request: _TRequest,
    ) -> _CallFuture[_TResponse]: ...

@type_check_only
class _CallIterator(Call, Generic[_TResponse], metaclass=abc.ABCMeta):
    def __iter__(self) -> Iterator[_TResponse]: ...
    def __next__(self) -> _TResponse: ...

class UnaryStreamClientInterceptor(abc.ABC):
    # This method (not the class) is generic over _TRequest and _TResponse
    # and the types must satisfy the no-op implementation of
    # `return continuation(client_call_details, request)`.
    @abc.abstractmethod
    def intercept_unary_stream(
        self,
        continuation: Callable[[ClientCallDetails, _TRequest], _CallIterator[_TResponse]],
        client_call_details: ClientCallDetails,
        request: _TRequest,
    ) -> _CallIterator[_TResponse]: ...

class StreamUnaryClientInterceptor(abc.ABC):
    # This method (not the class) is generic over _TRequest and _TResponse
    # and the types must satisfy the no-op implementation of
    # `return continuation(client_call_details, request_iterator)`.
    @abc.abstractmethod
    def intercept_stream_unary(
        self,
        continuation: Callable[[ClientCallDetails, Iterator[_TRequest]], _CallFuture[_TResponse]],
        client_call_details: ClientCallDetails,
        request_iterator: Iterator[_TRequest],
    ) -> _CallFuture[_TResponse]: ...

class StreamStreamClientInterceptor(abc.ABC):
    # This method (not the class) is generic over _TRequest and _TResponse
    # and the types must satisfy the no-op implementation of
    # `return continuation(client_call_details, request_iterator)`.
    @abc.abstractmethod
    def intercept_stream_stream(
        self,
        continuation: Callable[[ClientCallDetails, Iterator[_TRequest]], _CallIterator[_TResponse]],
        client_call_details: ClientCallDetails,
        request_iterator: Iterator[_TRequest],
    ) -> _CallIterator[_TResponse]: ...

# Service-Side Context:

class ServicerContext(RpcContext, metaclass=abc.ABCMeta):
    # misnamed parameter 'details', does not align with status.proto, where it is called 'message':
    @abc.abstractmethod
    def abort(self, code: StatusCode, details: str) -> NoReturn: ...
    @abc.abstractmethod
    def abort_with_status(self, status: Status) -> NoReturn: ...

    # FIXME: The docs say "A map of strings to an iterable of bytes for each auth property".
    # Does that mean 'bytes' (which is iterable), or 'Iterable[bytes]'?
    @abc.abstractmethod
    def auth_context(self) -> Mapping[str, bytes]: ...
    def disable_next_message_compression(self) -> None: ...
    @abc.abstractmethod
    def invocation_metadata(self) -> _Metadata: ...
    @abc.abstractmethod
    def peer(self) -> str: ...
    @abc.abstractmethod
    def peer_identities(self) -> Iterable[bytes] | None: ...
    @abc.abstractmethod
    def peer_identity_key(self) -> str | None: ...
    @abc.abstractmethod
    def send_initial_metadata(self, initial_metadata: _Metadata) -> None: ...
    @abc.abstractmethod
    def set_code(self, code: StatusCode) -> None: ...
    def set_compression(self, compression: Compression) -> None: ...
    @abc.abstractmethod
    def set_trailing_metadata(self, trailing_metadata: _Metadata) -> None: ...

    # misnamed function 'details', does not align with status.proto, where it is called 'message':
    @abc.abstractmethod
    def set_details(self, details: str) -> None: ...
    def trailing_metadata(self) -> _Metadata: ...

# Service-Side Handler:

class RpcMethodHandler(abc.ABC, Generic[_TRequest, _TResponse]):
    request_streaming: bool
    response_streaming: bool

    # XXX: not clear from docs whether this is optional or not
    request_deserializer: _Deserializer[_TRequest] | None

    # XXX: not clear from docs whether this is optional or not
    response_serializer: _Serializer[_TResponse] | None

    unary_unary: Callable[[_TRequest, ServicerContext], _TResponse] | None

    unary_stream: Callable[[_TRequest, ServicerContext], Iterator[_TResponse]] | None

    stream_unary: Callable[[Iterator[_TRequest], ServicerContext], _TResponse] | None

    stream_stream: Callable[[Iterator[_TRequest], ServicerContext], Iterator[_TResponse]] | None

class HandlerCallDetails(abc.ABC):
    method: str
    invocation_metadata: _Metadata

class GenericRpcHandler(abc.ABC):
    # The return type depends on the handler call details.
    @abc.abstractmethod
    def service(self, handler_call_details: HandlerCallDetails) -> RpcMethodHandler[Any, Any] | None: ...

class ServiceRpcHandler(GenericRpcHandler, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def service_name(self) -> str: ...

# Service-Side Interceptor:

class ServerInterceptor(abc.ABC):
    # This method (not the class) is generic over _TRequest and _TResponse
    # and the types must satisfy the no-op implementation of
    # `return continuation(handler_call_details)`.
    @abc.abstractmethod
    def intercept_service(
        self,
        continuation: Callable[[HandlerCallDetails], RpcMethodHandler[_TRequest, _TResponse] | None],
        handler_call_details: HandlerCallDetails,
    ) -> RpcMethodHandler[_TRequest, _TResponse] | None: ...

# Multi-Callable Interfaces:

class UnaryUnaryMultiCallable(abc.ABC, Generic[_TRequest, _TResponse]):
    @abc.abstractmethod
    def __call__(
        self,
        request: _TRequest,
        timeout: float | None = None,
        metadata: _Metadata | None = None,
        credentials: CallCredentials | None = None,
        wait_for_ready: bool | None = None,
        compression: Compression | None = None,
    ) -> _TResponse: ...
    @abc.abstractmethod
    def future(
        self,
        request: _TRequest,
        timeout: float | None = None,
        metadata: _Metadata | None = None,
        credentials: CallCredentials | None = None,
        wait_for_ready: bool | None = None,
        compression: Compression | None = None,
    ) -> _CallFuture[_TResponse]: ...
    @abc.abstractmethod
    def with_call(
        self,
        request: _TRequest,
        timeout: float | None = None,
        metadata: _Metadata | None = None,
        credentials: CallCredentials | None = None,
        wait_for_ready: bool | None = None,
        compression: Compression | None = None,
        # FIXME: Return value is documented as "The response value for the RPC and a Call value for the RPC";
        # this is slightly unclear so this return type is a best-effort guess.
    ) -> tuple[_TResponse, Call]: ...

class UnaryStreamMultiCallable(abc.ABC, Generic[_TRequest, _TResponse]):
    @abc.abstractmethod
    def __call__(
        self,
        request: _TRequest,
        timeout: float | None = None,
        metadata: _Metadata | None = None,
        credentials: CallCredentials | None = None,
        wait_for_ready: bool | None = None,
        compression: Compression | None = None,
    ) -> _CallIterator[_TResponse]: ...

class StreamUnaryMultiCallable(abc.ABC, Generic[_TRequest, _TResponse]):
    @abc.abstractmethod
    def __call__(
        self,
        request_iterator: Iterator[_TRequest],
        timeout: float | None = None,
        metadata: _Metadata | None = None,
        credentials: CallCredentials | None = None,
        wait_for_ready: bool | None = None,
        compression: Compression | None = None,
    ) -> _TResponse: ...
    @abc.abstractmethod
    def future(
        self,
        request_iterator: Iterator[_TRequest],
        timeout: float | None = None,
        metadata: _Metadata | None = None,
        credentials: CallCredentials | None = None,
        wait_for_ready: bool | None = None,
        compression: Compression | None = None,
    ) -> _CallFuture[_TResponse]: ...
    @abc.abstractmethod
    def with_call(
        self,
        request_iterator: Iterator[_TRequest],
        timeout: float | None = None,
        metadata: _Metadata | None = None,
        credentials: CallCredentials | None = None,
        wait_for_ready: bool | None = None,
        compression: Compression | None = None,
        # FIXME: Return value is documented as "The response value for the RPC and a Call value for the RPC";
        # this is slightly unclear so this return type is a best-effort guess.
    ) -> tuple[_TResponse, Call]: ...

class StreamStreamMultiCallable(abc.ABC, Generic[_TRequest, _TResponse]):
    @abc.abstractmethod
    def __call__(
        self,
        request_iterator: Iterator[_TRequest],
        timeout: float | None = None,
        metadata: _Metadata | None = None,
        credentials: CallCredentials | None = None,
        wait_for_ready: bool | None = None,
        compression: Compression | None = None,
    ) -> _CallIterator[_TResponse]: ...

# Runtime Protobuf Parsing:

def protos(protobuf_path: str) -> ModuleType: ...
def services(protobuf_path: str) -> ModuleType: ...
def protos_and_services(protobuf_path: str) -> tuple[ModuleType, ModuleType]: ...



from _typeshed import Incomplete, SupportsLenAndGetItem

from antlr4.atn.ATN import ATN as ATN
from antlr4.error.Errors import IllegalStateException as IllegalStateException
from antlr4.RuleContext import RuleContext as RuleContext

class PredictionContext:
    EMPTY: Incomplete
    EMPTY_RETURN_STATE: int
    globalNodeCount: int
    id = globalNodeCount
    cachedHashCode: Incomplete
    def __init__(self, cachedHashCode: int) -> None: ...
    def __len__(self) -> int: ...
    def isEmpty(self): ...
    def hasEmptyPath(self): ...
    def getReturnState(self, index: int): ...
    def __hash__(self): ...

def calculateHashCode(parent: PredictionContext, returnState: int): ...
def calculateListsHashCode(parents: list[PredictionContext], returnStates: list[int]): ...

class PredictionContextCache:
    cache: Incomplete
    def __init__(self) -> None: ...
    def add(self, ctx: PredictionContext): ...
    def get(self, ctx: PredictionContext): ...
    def __len__(self) -> int: ...

class SingletonPredictionContext(PredictionContext):
    @staticmethod
    def create(parent: PredictionContext, returnState: int): ...
    parentCtx: Incomplete
    returnState: Incomplete
    def __init__(self, parent: PredictionContext, returnState: int) -> None: ...
    def __len__(self) -> int: ...
    def getParent(self, index: int): ...
    def getReturnState(self, index: int): ...
    def __eq__(self, other): ...
    def __hash__(self): ...

class EmptyPredictionContext(SingletonPredictionContext):
    def __init__(self) -> None: ...
    def isEmpty(self): ...
    def __eq__(self, other): ...
    def __hash__(self): ...

class ArrayPredictionContext(PredictionContext):
    parents: Incomplete
    returnStates: Incomplete
    def __init__(self, parents: list[PredictionContext], returnStates: list[int]) -> None: ...
    def isEmpty(self): ...
    def __len__(self) -> int: ...
    def getParent(self, index: int): ...
    def getReturnState(self, index: int): ...
    def __eq__(self, other): ...
    def __hash__(self): ...

def PredictionContextFromRuleContext(atn: ATN, outerContext: RuleContext | None = None): ...
def merge(
    a: PredictionContext,
    b: PredictionContext,
    rootIsWildcard: bool,
    mergeCache: dict[tuple[Incomplete, Incomplete], SingletonPredictionContext] | None,
): ...
def mergeSingletons(
    a: SingletonPredictionContext,
    b: SingletonPredictionContext,
    rootIsWildcard: bool,
    mergeCache: dict[tuple[Incomplete, Incomplete], SingletonPredictionContext] | None,
): ...
def mergeRoot(a: SingletonPredictionContext, b: SingletonPredictionContext, rootIsWildcard: bool): ...
def mergeArrays(
    a: ArrayPredictionContext,
    b: ArrayPredictionContext,
    rootIsWildcard: bool,
    mergeCache: dict[tuple[Incomplete, Incomplete], SingletonPredictionContext] | None,
): ...
def combineCommonParents(parents: SupportsLenAndGetItem[PredictionContext]): ...
def getCachedPredictionContext(
    context: PredictionContext, contextCache: PredictionContextCache, visited: dict[PredictionContext, PredictionContext]
): ...
def getAllContextNodes(
    context: PredictionContext,
    nodes: list[Incomplete] | None = None,
    visited: dict[PredictionContext, PredictionContext] | None = None,
): ...



import inspect
from builtins import dict as _dict  # alias to avoid conflicts with attribute name
from collections.abc import Callable, Generator, Iterator
from contextlib import _GeneratorContextManager
from inspect import Signature, getfullargspec as getfullargspec, iscoroutinefunction as iscoroutinefunction
from re import Pattern
from typing import Any, Final, Literal, TypeVar
from typing_extensions import ParamSpec

_C = TypeVar("_C", bound=Callable[..., Any])
_Func = TypeVar("_Func", bound=Callable[..., Any])
_T = TypeVar("_T")
_P = ParamSpec("_P")

DEF: Final[Pattern[str]]
POS: Final[Literal[inspect._ParameterKind.POSITIONAL_OR_KEYWORD]]
EMPTY: Final[type[inspect._empty]]

class FunctionMaker:
    args: list[str]
    varargs: str | None
    varkw: str | None
    defaults: tuple[Any, ...] | None
    kwonlyargs: list[str]
    kwonlydefaults: _dict[str, Any] | None
    shortsignature: str | None
    name: str
    doc: str | None
    module: str | None
    annotations: _dict[str, Any]
    signature: str
    dict: _dict[str, Any]
    def __init__(
        self,
        func: Callable[..., Any] | None = ...,
        name: str | None = ...,
        signature: str | None = ...,
        defaults: tuple[Any, ...] | None = ...,
        doc: str | None = ...,
        module: str | None = ...,
        funcdict: _dict[str, Any] | None = ...,
    ) -> None: ...
    def update(self, func: Any, **kw: Any) -> None: ...
    def make(
        self, src_templ: str, evaldict: _dict[str, Any] | None = ..., addsource: bool = ..., **attrs: Any
    ) -> Callable[..., Any]: ...
    @classmethod
    def create(
        cls,
        obj: Any,
        body: str,
        evaldict: _dict[str, Any],
        defaults: tuple[Any, ...] | None = ...,
        doc: str | None = ...,
        module: str | None = ...,
        addsource: bool = ...,
        **attrs: Any,
    ) -> Callable[..., Any]: ...

def fix(args: tuple[Any, ...], kwargs: dict[str, Any], sig: Signature) -> tuple[tuple[Any, ...], dict[str, Any]]: ...
def decorate(func: _Func, caller: Callable[..., Any], extras: tuple[Any, ...] = ..., kwsyntax: bool = False) -> _Func: ...
def decoratorx(caller: Callable[..., Any]) -> Callable[..., Any]: ...
def decorator(
    caller: Callable[..., Any], _func: Callable[..., Any] | None = None, kwsyntax: bool = False
) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...

class ContextManager(_GeneratorContextManager[_T]):
    def __init__(self, g: Callable[..., Generator[_T]], *a: Any, **k: Any) -> None: ...
    def __call__(self, func: _C) -> _C: ...

def contextmanager(func: Callable[_P, Iterator[_T]]) -> Callable[_P, ContextManager[_T]]: ...
def append(a: type, vancestors: list[type]) -> None: ...
def dispatch_on(*dispatch_args: Any) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...



############################################################################
# Copyright(c) Open Law Library. All rights reserved.                      #
# See ThirdPartyNotices.txt in the project root for additional notices.    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License")           #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#     http: // www.apache.org/licenses/LICENSE-2.0                         #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
############################################################################
import asyncio
import json
import logging
import re
import sys
from concurrent.futures import Future, ThreadPoolExecutor
from threading import Event
from typing import (
    Any,
    Callable,
    List,
    Optional,
    TextIO,
    Type,
    TypeVar,
    Union,
)

import cattrs
from pygls import IS_PYODIDE
from pygls.lsp import ConfigCallbackType, ShowDocumentCallbackType
from pygls.exceptions import (
    FeatureNotificationError,
    JsonRpcInternalError,
    PyglsError,
    JsonRpcException,
    FeatureRequestError,
)
from lsprotocol.types import (
    ClientCapabilities,
    Diagnostic,
    MessageType,
    NotebookDocumentSyncOptions,
    RegistrationParams,
    ServerCapabilities,
    ShowDocumentParams,
    TextDocumentSyncKind,
    UnregistrationParams,
    WorkspaceApplyEditResponse,
    WorkspaceEdit,
    WorkspaceConfigurationParams,
)
from pygls.progress import Progress
from pygls.protocol import JsonRPCProtocol, LanguageServerProtocol, default_converter
from pygls.workspace import Workspace

if not IS_PYODIDE:
    from multiprocessing.pool import ThreadPool


logger = logging.getLogger(__name__)

F = TypeVar("F", bound=Callable)

ServerErrors = Union[
    PyglsError,
    JsonRpcException,
    Type[JsonRpcInternalError],
    Type[FeatureNotificationError],
    Type[FeatureRequestError],
]


async def aio_readline(loop, executor, stop_event, rfile, proxy):
    """Reads data from stdin in separate thread (asynchronously)."""

    CONTENT_LENGTH_PATTERN = re.compile(rb"^Content-Length: (\d+)\r\n$")

    # Initialize message buffer
    message = []
    content_length = 0

    while not stop_event.is_set() and not rfile.closed:
        # Read a header line
        header = await loop.run_in_executor(executor, rfile.readline)
        if not header:
            break
        message.append(header)

        # Extract content length if possible
        if not content_length:
            match = CONTENT_LENGTH_PATTERN.fullmatch(header)
            if match:
                content_length = int(match.group(1))
                logger.debug("Content length: %s", content_length)

        # Check if all headers have been read (as indicated by an empty line \r\n)
        if content_length and not header.strip():
            # Read body
            body = await loop.run_in_executor(executor, rfile.read, content_length)
            if not body:
                break
            message.append(body)

            # Pass message to language server protocol
            proxy(b"".join(message))

            # Reset the buffer
            message = []
            content_length = 0


class StdOutTransportAdapter:
    """Protocol adapter which overrides write method.

    Write method sends data to stdout.
    """

    def __init__(self, rfile, wfile):
        self.rfile = rfile
        self.wfile = wfile

    def close(self):
        self.rfile.close()
        self.wfile.close()

    def write(self, data):
        self.wfile.write(data)
        self.wfile.flush()


class PyodideTransportAdapter:
    """Protocol adapter which overrides write method.

    Write method sends data to stdout.
    """

    def __init__(self, wfile):
        self.wfile = wfile

    def close(self):
        self.wfile.close()

    def write(self, data):
        self.wfile.write(data)
        self.wfile.flush()


class WebSocketTransportAdapter:
    """Protocol adapter which calls write method.

    Write method sends data via the WebSocket interface.
    """

    def __init__(self, ws, loop):
        self._ws = ws
        self._loop = loop

    def close(self) -> None:
        """Stop the WebSocket server."""
        self._ws.close()

    def write(self, data: Any) -> None:
        """Create a task to write specified data into a WebSocket."""
        asyncio.ensure_future(self._ws.send(data))


class Server:
    """Base server class

    Parameters
    ----------
    protocol_cls
       Protocol implementation that must be derive from :class:`~pygls.protocol.JsonRPCProtocol`

    converter_factory
       Factory function to use when constructing a cattrs converter.

    loop
       The asyncio event loop

    max_workers
       Maximum number of workers for `ThreadPool` and `ThreadPoolExecutor`

    """

    def __init__(
        self,
        protocol_cls: Type[JsonRPCProtocol],
        converter_factory: Callable[[], cattrs.Converter],
        loop: Optional[asyncio.AbstractEventLoop] = None,
        max_workers: int = 2,
        sync_kind: TextDocumentSyncKind = TextDocumentSyncKind.Incremental,
    ):
        if not issubclass(protocol_cls, asyncio.Protocol):
            raise TypeError("Protocol class should be subclass of asyncio.Protocol")

        self._max_workers = max_workers
        self._server = None
        self._stop_event: Optional[Event] = None
        self._thread_pool: Optional[ThreadPool] = None
        self._thread_pool_executor: Optional[ThreadPoolExecutor] = None

        if sync_kind is not None:
            self.text_document_sync_kind = sync_kind

        if loop is None:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            self._owns_loop = True
        else:
            self._owns_loop = False

        self.loop = loop

        # TODO: Will move this to `LanguageServer` soon
        self.lsp = protocol_cls(self, converter_factory())  # type: ignore

    def shutdown(self):
        """Shutdown server."""
        logger.info("Shutting down the server")

        if self._stop_event is not None:
            self._stop_event.set()

        if self._thread_pool:
            self._thread_pool.terminate()
            self._thread_pool.join()

        if self._thread_pool_executor:
            self._thread_pool_executor.shutdown()

        if self._server:
            self._server.close()
            self.loop.run_until_complete(self._server.wait_closed())

        if self._owns_loop and not self.loop.is_closed():
            logger.info("Closing the event loop.")
            self.loop.close()

    def start_io(self, stdin: Optional[TextIO] = None, stdout: Optional[TextIO] = None):
        """Starts IO server."""
        logger.info("Starting IO server")

        self._stop_event = Event()
        transport = StdOutTransportAdapter(
            stdin or sys.stdin.buffer, stdout or sys.stdout.buffer
        )
        self.lsp.connection_made(transport)  # type: ignore[arg-type]

        try:
            self.loop.run_until_complete(
                aio_readline(
                    self.loop,
                    self.thread_pool_executor,
                    self._stop_event,
                    stdin or sys.stdin.buffer,
                    self.lsp.data_received,
                )
            )
        except BrokenPipeError:
            logger.error("Connection to the client is lost! Shutting down the server.")
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            self.shutdown()

    def start_pyodide(self):
        logger.info("Starting Pyodide server")

        # Note: We don't actually start anything running as the main event
        # loop will be handled by the web platform.
        transport = PyodideTransportAdapter(sys.stdout)
        self.lsp.connection_made(transport)  # type: ignore[arg-type]
        self.lsp._send_only_body = True  # Don't send headers within the payload

    def start_tcp(self, host: str, port: int) -> None:
        """Starts TCP server."""
        logger.info("Starting TCP server on %s:%s", host, port)

        self._stop_event = Event()
        self._server = self.loop.run_until_complete(  # type: ignore[assignment]
            self.loop.create_server(self.lsp, host, port)
        )
        try:
            self.loop.run_forever()
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            self.shutdown()

    def start_ws(self, host: str, port: int) -> None:
        """Starts WebSocket server."""
        try:
            from websockets.server import serve
        except ImportError:
            logger.error("Run `pip install pygls[ws]` to install `websockets`.")
            sys.exit(1)

        logger.info("Starting WebSocket server on {}:{}".format(host, port))

        self._stop_event = Event()
        self.lsp._send_only_body = True  # Don't send headers within the payload

        async def connection_made(websocket, _):
            """Handle new connection wrapped in the WebSocket."""
            self.lsp.transport = WebSocketTransportAdapter(websocket, self.loop)
            async for message in websocket:
                self.lsp._procedure_handler(
                    json.loads(message, object_hook=self.lsp._deserialize_message)
                )

        start_server = serve(connection_made, host, port, loop=self.loop)
        self._server = start_server.ws_server  # type: ignore[assignment]
        self.loop.run_until_complete(start_server)

        try:
            self.loop.run_forever()
        except (KeyboardInterrupt, SystemExit):
            pass
        finally:
            self._stop_event.set()
            self.shutdown()

    if not IS_PYODIDE:

        @property
        def thread_pool(self) -> ThreadPool:
            """Returns thread pool instance (lazy initialization)."""
            if not self._thread_pool:
                self._thread_pool = ThreadPool(processes=self._max_workers)

            return self._thread_pool

        @property
        def thread_pool_executor(self) -> ThreadPoolExecutor:
            """Returns thread pool instance (lazy initialization)."""
            if not self._thread_pool_executor:
                self._thread_pool_executor = ThreadPoolExecutor(
                    max_workers=self._max_workers
                )

            return self._thread_pool_executor


class LanguageServer(Server):
    """The default LanguageServer

    This class can be extended and it can be passed as a first argument to
    registered commands/features.

    .. |ServerInfo| replace:: :class:`~lsprotocol.types.InitializeResultServerInfoType`

    Parameters
    ----------
    name
       Name of the server, used to populate |ServerInfo| which is sent to
       the client during initialization

    version
       Version of the server, used to populate |ServerInfo| which is sent to
       the client during initialization

    protocol_cls
       The :class:`~pygls.protocol.LanguageServerProtocol` class definition, or any
       subclass of it.

    max_workers
       Maximum number of workers for ``ThreadPool`` and ``ThreadPoolExecutor``

    text_document_sync_kind
       Text document synchronization method

       None
          No synchronization

       :attr:`~lsprotocol.types.TextDocumentSyncKind.Full`
          Send entire document text with each update

       :attr:`~lsprotocol.types.TextDocumentSyncKind.Incremental`
          Send only the region of text that changed with each update

    notebook_document_sync
       Advertise :lsp:`NotebookDocument` support to the client.
    """

    lsp: LanguageServerProtocol

    default_error_message = (
        "Unexpected error in LSP server, see server's logs for details"
    )
    """
    The default error message sent to the user's editor when this server encounters an uncaught
    exception.
    """

    def __init__(
        self,
        name: str,
        version: str,
        loop=None,
        protocol_cls: Type[LanguageServerProtocol] = LanguageServerProtocol,
        converter_factory=default_converter,
        text_document_sync_kind: TextDocumentSyncKind = TextDocumentSyncKind.Incremental,
        notebook_document_sync: Optional[NotebookDocumentSyncOptions] = None,
        max_workers: int = 2,
    ):
        if not issubclass(protocol_cls, LanguageServerProtocol):
            raise TypeError(
                "Protocol class should be subclass of LanguageServerProtocol"
            )

        self.name = name
        self.version = version
        self._text_document_sync_kind = text_document_sync_kind
        self._notebook_document_sync = notebook_document_sync
        self.process_id: Optional[Union[int, None]] = None
        super().__init__(protocol_cls, converter_factory, loop, max_workers)

    def apply_edit(
        self, edit: WorkspaceEdit, label: Optional[str] = None
    ) -> WorkspaceApplyEditResponse:
        """Sends apply edit request to the client."""
        return self.lsp.apply_edit(edit, label)

    def apply_edit_async(
        self, edit: WorkspaceEdit, label: Optional[str] = None
    ) -> WorkspaceApplyEditResponse:
        """Sends apply edit request to the client. Should be called with `await`"""
        return self.lsp.apply_edit_async(edit, label)

    def command(self, command_name: str) -> Callable[[F], F]:
        """Decorator used to register custom commands.

        Example
        -------
        ::

           @ls.command('myCustomCommand')
           def my_cmd(ls, a, b, c):
               pass
        """
        return self.lsp.fm.command(command_name)

    @property
    def client_capabilities(self) -> ClientCapabilities:
        """The client's capabilities."""
        return self.lsp.client_capabilities

    def feature(
        self,
        feature_name: str,
        options: Optional[Any] = None,
    ) -> Callable[[F], F]:
        """Decorator used to register LSP features.

        Example
        -------
        ::

           @ls.feature('textDocument/completion', CompletionOptions(trigger_characters=['.']))
           def completions(ls, params: CompletionParams):
               return CompletionList(is_incomplete=False, items=[CompletionItem("Completion 1")])
        """
        return self.lsp.fm.feature(feature_name, options)

    def get_configuration(
        self,
        params: WorkspaceConfigurationParams,
        callback: Optional[ConfigCallbackType] = None,
    ) -> Future:
        """Gets the configuration settings from the client."""
        return self.lsp.get_configuration(params, callback)

    def get_configuration_async(
        self, params: WorkspaceConfigurationParams
    ) -> asyncio.Future:
        """Gets the configuration settings from the client. Should be called with `await`"""
        return self.lsp.get_configuration_async(params)

    def log_trace(self, message: str, verbose: Optional[str] = None) -> None:
        """Sends trace notification to the client."""
        self.lsp.log_trace(message, verbose)

    @property
    def progress(self) -> Progress:
        """Gets the object to manage client's progress bar."""
        return self.lsp.progress

    def publish_diagnostics(
        self,
        uri: str,
        diagnostics: Optional[List[Diagnostic]] = None,
        version: Optional[int] = None,
        **kwargs
    ):
        """
        Sends diagnostic notification to the client.
        """
        params = self.lsp._construct_publish_diagnostic_type(
            uri, diagnostics, version, **kwargs
        )
        self.lsp.publish_diagnostics(params, **kwargs)

    def register_capability(
        self, params: RegistrationParams, callback: Optional[Callable[[], None]] = None
    ) -> Future:
        """Register a new capability on the client."""
        return self.lsp.register_capability(params, callback)

    def register_capability_async(self, params: RegistrationParams) -> asyncio.Future:
        """Register a new capability on the client. Should be called with `await`"""
        return self.lsp.register_capability_async(params)

    def semantic_tokens_refresh(
        self, callback: Optional[Callable[[], None]] = None
    ) -> Future:
        """Request a refresh of all semantic tokens."""
        return self.lsp.semantic_tokens_refresh(callback)

    def semantic_tokens_refresh_async(self) -> asyncio.Future:
        """Request a refresh of all semantic tokens. Should be called with `await`"""
        return self.lsp.semantic_tokens_refresh_async()

    def send_notification(self, method: str, params: object = None) -> None:
        """Sends notification to the client."""
        self.lsp.notify(method, params)

    @property
    def server_capabilities(self) -> ServerCapabilities:
        """Return server capabilities."""
        return self.lsp.server_capabilities

    def show_document(
        self,
        params: ShowDocumentParams,
        callback: Optional[ShowDocumentCallbackType] = None,
    ) -> Future:
        """Display a particular document in the user interface."""
        return self.lsp.show_document(params, callback)

    def show_document_async(self, params: ShowDocumentParams) -> asyncio.Future:
        """Display a particular document in the user interface. Should be called with `await`"""
        return self.lsp.show_document_async(params)

    def show_message(self, message, msg_type=MessageType.Info) -> None:
        """Sends message to the client to display message."""
        self.lsp.show_message(message, msg_type)

    def show_message_log(self, message, msg_type=MessageType.Log) -> None:
        """Sends message to the client's output channel."""
        self.lsp.show_message_log(message, msg_type)

    def _report_server_error(
        self,
        error: Exception,
        source: ServerErrors,
    ):
        # Prevent recursive error reporting
        try:
            self.report_server_error(error, source)
        except Exception:
            logger.warning("Failed to report error to client")

    def report_server_error(self, error: Exception, source: ServerErrors):
        """
        Sends error to the client for displaying.

        By default this fucntion does not handle LSP request errors. This is because LSP requests
        require direct responses and so already have a mechanism for including unexpected errors
        in the response body.

        All other errors are "out of band" in the sense that the client isn't explicitly waiting
        for them. For example diagnostics are returned as notifications, not responses to requests,
        and so can seemingly be sent at random. Also for example consider JSON RPC serialization
        and deserialization, if a payload cannot be parsed then the whole request/response cycle
        cannot be completed and so one of these "out of band" error messages is sent.

        These "out of band" error messages are not a requirement of the LSP spec. Pygls simply
        offers this behaviour as a recommended default. It is perfectly reasonble to override this
        default.
        """

        if source == FeatureRequestError:
            return

        self.show_message(self.default_error_message, msg_type=MessageType.Error)

    def thread(self) -> Callable[[F], F]:
        """Decorator that mark function to execute it in a thread."""
        return self.lsp.thread()

    def unregister_capability(
        self,
        params: UnregistrationParams,
        callback: Optional[Callable[[], None]] = None,
    ) -> Future:
        """Unregister a new capability on the client."""
        return self.lsp.unregister_capability(params, callback)

    def unregister_capability_async(
        self, params: UnregistrationParams
    ) -> asyncio.Future:
        """Unregister a new capability on the client. Should be called with `await`"""
        return self.lsp.unregister_capability_async(params)

    @property
    def workspace(self) -> Workspace:
        """Returns in-memory workspace."""
        return self.lsp.workspace



# Type aliases used in this stub package
from _typeshed import SupportsWrite
from typing import Any, Protocol
from typing_extensions import TypeAlias

Box: TypeAlias = tuple[tuple[int, int], tuple[int, int]]
Ink: TypeAlias = tuple[int, int, int] | tuple[int, int, int, int]

# Don't try to make these Literal[x, y, z] as this really wreaks
# havoc with overloads in mypy.
ErrorCorrect: TypeAlias = int
MaskPattern: TypeAlias = int

class Writeable(SupportsWrite[bytes], Protocol):
    def seek(self, offset: int, /) -> Any: ...



from _typeshed import Unused
from collections.abc import Iterable
from typing import TypedDict, type_check_only
from typing_extensions import NotRequired

import gdb

from ..FrameDecorator import FrameDecorator, SymValueWrapper
from .varref import BaseReference, _ReferenceDescriptor

frame_to_scope: dict[int, _ScopeReference]

@type_check_only
class _ScopeReferenceDescriptor(_ReferenceDescriptor):
    presentationHint: str
    expensive: bool
    namedVariables: int
    line: NotRequired[int]

@type_check_only
class _ScopesResult(TypedDict):
    scopes: list[_ScopeReferenceDescriptor]

def clear_scopes(event: Unused) -> None: ...
def set_finish_value(val: gdb.Value) -> None: ...
def symbol_value(sym: SymValueWrapper, frame: FrameDecorator) -> tuple[str, gdb.Value]: ...

class _ScopeReference(BaseReference):
    hint: str
    frame: FrameDecorator
    inf_frame: gdb.Frame
    function: str | None
    line: int | None
    var_list: tuple[SymValueWrapper, ...]
    def __init__(self, name: str, hint: str, frame: FrameDecorator, var_list: Iterable[SymValueWrapper]) -> None: ...
    def to_object(self) -> _ScopeReferenceDescriptor: ...
    def has_children(self) -> bool: ...
    def child_count(self) -> int: ...
    # note: parameter named changed from 'index' to 'idx'
    def fetch_one_child(self, idx: int) -> tuple[str, gdb.Value]: ...

class _FinishScopeReference(_ScopeReference): ...

class _RegisterReference(_ScopeReference):
    def __init__(self, name: str, frame: FrameDecorator) -> None: ...

def scopes(*, frameId: int, **extra: Unused) -> _ScopesResult: ...



from braintree.attribute_getter import AttributeGetter
from braintree.sub_merchant import SubMerchant

class PaymentFacilitator(AttributeGetter):
    sub_merchant: SubMerchant
    def __init__(self, attributes) -> None: ...



from typing import Literal

import numpy as np
from networkx._typing import Array1D, Seed
from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["algebraic_connectivity", "fiedler_vector", "spectral_ordering", "spectral_bisection"]

@_dispatchable
def algebraic_connectivity(
    G: Graph[_Node],
    weight: str | None = "weight",
    normalized: bool = False,
    tol: float = 1e-08,
    method: Literal["tracemin_pcg", "tracemin_lu", "lanczos", "lobpcg"] = "tracemin_pcg",
    seed: Seed | None = None,
) -> float: ...
@_dispatchable
def fiedler_vector(
    G: Graph[_Node],
    weight: str | None = "weight",
    normalized: bool = False,
    tol: float = 1e-08,
    method: Literal["tracemin_pcg", "tracemin_lu", "lanczos", "lobpcg"] = "tracemin_pcg",
    seed: Seed | None = None,
) -> Array1D[np.float64]: ...
@_dispatchable
def spectral_ordering(
    G: Graph[_Node],
    weight: str | None = "weight",
    normalized: bool = False,
    tol: float = 1e-08,
    method: Literal["tracemin_pcg", "tracemin_lu", "lanczos", "lobpcg"] = "tracemin_pcg",
    seed: Seed | None = None,
) -> list[_Node]: ...
@_dispatchable
def spectral_bisection(
    G: Graph[_Node],
    weight: str | None = "weight",
    normalized: bool = False,
    tol: float = 1e-08,
    method: Literal["tracemin_pcg", "tracemin_lu", "lanczos", "lobpcg"] = "tracemin_pcg",
    seed: Seed | None = None,
) -> tuple[set[_Node], set[_Node]]: ...



import sys
from typing_extensions import Self, TypeAlias, disjoint_base

# Actually Tuple[(int,) * 625]
_State: TypeAlias = tuple[int, ...]

@disjoint_base
class Random:
    if sys.version_info >= (3, 10):
        def __init__(self, seed: object = ..., /) -> None: ...
    else:
        def __new__(self, seed: object = ..., /) -> Self: ...

    def seed(self, n: object = None, /) -> None: ...
    def getstate(self) -> _State: ...
    def setstate(self, state: _State, /) -> None: ...
    def random(self) -> float: ...
    def getrandbits(self, k: int, /) -> int: ...



from hvac.api.vault_api_base import VaultApiBase

class Aws(VaultApiBase):
    def configure_root_iam_credentials(
        self, access_key, secret_key, region=None, iam_endpoint=None, sts_endpoint=None, max_retries=None, mount_point="aws"
    ): ...
    def rotate_root_iam_credentials(self, mount_point="aws"): ...
    def configure_lease(self, lease, lease_max, mount_point="aws"): ...
    def read_lease_config(self, mount_point="aws"): ...
    def create_or_update_role(
        self,
        name,
        credential_type,
        policy_document=None,
        default_sts_ttl=None,
        max_sts_ttl=None,
        role_arns=None,
        policy_arns=None,
        legacy_params: bool = False,
        iam_tags=None,
        mount_point="aws",
    ): ...
    def read_role(self, name, mount_point="aws"): ...
    def list_roles(self, mount_point="aws"): ...
    def delete_role(self, name, mount_point="aws"): ...
    def generate_credentials(
        self, name, role_arn=None, ttl=None, endpoint: str = "creds", mount_point="aws", role_session_name=None
    ): ...



from .std import TqdmDeprecationWarning as TqdmDeprecationWarning
from .utils import (
    CUR_OS as CUR_OS,
    IS_NIX as IS_NIX,
    IS_WIN as IS_WIN,
    RE_ANSI as RE_ANSI,
    Comparable as Comparable,
    FormatReplace as FormatReplace,
    SimpleTextIOWrapper as SimpleTextIOWrapper,
)
